@@ -5,8 +5,7 @@ import shutil
 from collections import OrderedDict
 from typing import List, Dict, Tuple, Iterable, Type, Union
 from zipfile import ZipFile
-import time
-from multiprocessing import Pool, cpu_count
+import requests
 import numpy as np
 import transformers
 import torch
@@ -15,27 +14,29 @@ from torch import nn, Tensor, device
 from torch.optim import Optimizer
 from torch.utils.data import DataLoader
 from tqdm.autonotebook import tqdm, trange
-import multiprocessing
+import torch.multiprocessing as mp
+import math
+import queue
 
 from . import __DOWNLOAD_SERVER__
 from .evaluation import SentenceEvaluator
 from .util import import_from_string, batch_to_device, http_get
+from .datasets.EncodeDataset import EncodeDataset
+from .models import Transformer, Pooling
 from . import __version__
 
 class SentenceTransformer(nn.Sequential):
     def __init__(self, model_name_or_path: str = None, modules: Iterable[nn.Module] = None, device: str = None):
-        if modules is not None and not isinstance(modules, OrderedDict):
-            modules = OrderedDict([(str(idx), module) for idx, module in enumerate(modules)])
-
         if model_name_or_path is not None and model_name_or_path != "":
             logging.info("Load pretrained SentenceTransformer: {}".format(model_name_or_path))
+            model_path = model_name_or_path
 
-            if '/' not in model_name_or_path and '\\' not in model_name_or_path and not os.path.isdir(model_name_or_path):
-                logging.info("Did not find a '/' or '\\' in the name. Assume to download model from server.")
-                model_name_or_path = __DOWNLOAD_SERVER__ + model_name_or_path + '.zip'
+            if not os.path.isdir(model_path):
+                logging.info("Did not find folder {}. Assume to download model from server.".format(model_path))
+                model_path = __DOWNLOAD_SERVER__ + model_path + '.zip'
 
-            if model_name_or_path.startswith('http://') or model_name_or_path.startswith('https://'):
-                model_url = model_name_or_path
+            if model_path.startswith('http://') or model_path.startswith('https://'):
+                model_url = model_path
                 folder_name = model_url.replace("https://", "").replace("http://", "").replace("/", "_")[:250]
 
                 try:
@@ -58,11 +59,25 @@ class SentenceTransformer(nn.Sequential):
                         http_get(model_url, zip_save_path)
                         with ZipFile(zip_save_path, 'r') as zip:
                             zip.extractall(model_path)
+                    except requests.exceptions.HTTPError as e:
+                        shutil.rmtree(model_path)
+                        if e.response.status_code == 404:
+                            logging.warning('SentenceTransformer-Model {} not found. Try to create it from scratch'.format(model_url))
+                            logging.warning('Try to create Transformer Model {} with mean pooling'.format(model_name_or_path))
+
+                            model_path = None
+                            transformer_model = Transformer(model_name_or_path)
+                            pooling_model = Pooling(transformer_model.get_word_embedding_dimension())
+                            modules = [transformer_model, pooling_model]
+
+                        else:
+                            raise e
                     except Exception as e:
                         shutil.rmtree(model_path)
                         raise e
-            else:
-                model_path = model_name_or_path
+
+
+
 
             #### Load from disk
             if model_path is not None:
@@ -84,104 +99,69 @@ class SentenceTransformer(nn.Sequential):
                     modules[module_config['name']] = module
 
 
+        if modules is not None and not isinstance(modules, OrderedDict):
+            modules = OrderedDict([(str(idx), module) for idx, module in enumerate(modules)])
+
         super().__init__(modules)
         if device is None:
             device = "cuda" if torch.cuda.is_available() else "cpu"
             logging.info("Use pytorch device: {}".format(device))
 
         self._target_device = torch.device(device)
-        self.parallel_tokenization = multiprocessing.get_start_method() == 'fork'   #parallel_tokenization only works if the Operating System support fork
-        self.parallel_tokenization_processes = min(4, cpu_count())                  #Number of parallel processes used for tokenization. Increase up to cpu_count() for faster tokenization
-        self.parallel_tokenization_chunksize = 5000                                 #Number of sentences sent per chunk to each process. Increase for faster tokenization
+
 
     def encode(self, sentences: Union[str, List[str], List[int]],
-               batch_size: int = 8,
+               batch_size: int = 32,
                show_progress_bar: bool = None,
                output_value: str = 'sentence_embedding',
                convert_to_numpy: bool = True,
                convert_to_tensor: bool = False,
-               is_pretokenized: bool = False) -> Union[List[Tensor], ndarray, Tensor]:
+               is_pretokenized: bool = False,
+               device: str = None,
+               num_workers: int = 0) -> Union[List[Tensor], ndarray, Tensor]:
         """
         Computes sentence embeddings
 
-        :param sentences:
-           the sentences to embed
-        :param batch_size:
-           the batch size used for the computation
-        :param show_progress_bar:
-            Output a progress bar when encode sentences
-        :param output_value:
-            Default sentence_embedding, to get sentence embeddings. Can be set to token_embeddings
-            to get wordpiece token embeddings.
-        :param convert_to_numpy:
-            If true, the output is a list of numpy vectors. Else, it is a list of pytorch tensors.
-        :param convert_to_tensor:
-            If true, you get one large tensor as return. Overwrites any setting from conver_to_numy
-        :param is_pretokenized:
-            If is_pretokenized=True, sentences must be a list of integers, containing the tokenized sentences with each token convert to the respective int.
+
+        :param sentences: the sentences to embed
+        :param batch_size: the batch size used for the computation
+        :param show_progress_bar: Output a progress bar when encode sentences
+        :param output_value:  Default sentence_embedding, to get sentence embeddings. Can be set to token_embeddings to get wordpiece token embeddings.
+        :param convert_to_numpy: If true, the output is a list of numpy vectors. Else, it is a list of pytorch tensors.
+        :param convert_to_tensor: If true, you get one large tensor as return. Overwrites any setting from conver_to_numy
+        :param is_pretokenized: If is_pretokenized=True, sentences must be a list of integers, containing the tokenized sentences with each token convert to the respective int.
+        :param device: Which torch.device to use for the computation
+        :param num_workers: Number of background-workers to tokenize data. Set to positive number to increase tokenization speed
         :return:
            By default, a list of tensors is returned. If convert_to_tensor, a stacked tensor is returned. If convert_to_numpy, a numpy matrix is returned.
         """
-
         self.eval()
         if show_progress_bar is None:
             show_progress_bar = (logging.getLogger().getEffectiveLevel()==logging.INFO or logging.getLogger().getEffectiveLevel()==logging.DEBUG)
 
-
+        input_was_string = False
         if isinstance(sentences, str): #Cast an individual sentence to a list with length 1
             sentences = [sentences]
+            input_was_string = True
 
+        if device is None:
+            device = self._target_device
+
+        self.to(device)
 
         all_embeddings = []
-
-        if show_progress_bar:
-            logging.info("Start tokenization {} sentences".format(len(sentences)))
-
-        if is_pretokenized:
-            sentences_tokenized = sentences
-        else:
-            if not self.parallel_tokenization or len(sentences) < self.parallel_tokenization_chunksize:
-                sentences_tokenized = [self.tokenize(sen) for sen in sentences]
-            else:
-                if show_progress_bar:
-                    logging.info("Multi-process tokenization with {} workers".format(self.parallel_tokenization_processes))
-
-                self.to('cpu')   #Model must be on CPU to work with fork
-                with Pool(self.parallel_tokenization_processes) as p:
-                    sentences_tokenized = list(p.imap(self.tokenize, sentences, chunksize=self.parallel_tokenization_chunksize))
-
-        self.to(self._target_device)
-
         length_sorted_idx = np.argsort([len(sen) for sen in sentences])
+        sentences_sorted = [sentences[idx] for idx in length_sorted_idx]
+        inp_dataset = EncodeDataset(sentences_sorted, model=self, is_tokenized=is_pretokenized)
+        inp_dataloader = DataLoader(inp_dataset, batch_size=batch_size, collate_fn=self.smart_batching_collate_text_only, num_workers=num_workers, shuffle=False)
 
-        iterator = range(0, len(sentences), batch_size)
+        iterator = inp_dataloader
         if show_progress_bar:
-            iterator = tqdm(iterator, desc="Batches")
-
-        for batch_idx in iterator:
-            batch_tokens = []
-
-            batch_start = batch_idx
-            batch_end = min(batch_start + batch_size, len(sentences))
-
-            longest_seq = 0
-
-            for idx in length_sorted_idx[batch_start: batch_end]:
-                tokens = sentences_tokenized[idx]
-                longest_seq = max(longest_seq, len(tokens))
-                batch_tokens.append(tokens)
-
-            features = {}
-            for text in batch_tokens:
-                sentence_features = self.get_sentence_features(text, longest_seq)
-
-                for feature_name in sentence_features:
-                    if feature_name not in features:
-                        features[feature_name] = []
-                    features[feature_name].append(sentence_features[feature_name])
+            iterator = tqdm(inp_dataloader, desc="Batches")
 
+        for features in iterator:
             for feature_name in features:
-                features[feature_name] = torch.cat(features[feature_name]).to(self._target_device)
+                features[feature_name] = features[feature_name].to(device)
 
             with torch.no_grad():
                 out_features = self.forward(features)
@@ -195,24 +175,136 @@ class SentenceTransformer(nn.Sequential):
 
                 all_embeddings.extend(embeddings)
 
-        reverting_order = np.argsort(length_sorted_idx)
-        all_embeddings = [all_embeddings[idx] for idx in reverting_order]
+
+        all_embeddings = [all_embeddings[idx] for idx in np.argsort(length_sorted_idx)]
 
         if convert_to_tensor:
             all_embeddings = torch.stack(all_embeddings)
         elif convert_to_numpy:
             all_embeddings = np.asarray([emb.cpu().detach().numpy() for emb in all_embeddings])
 
+        if input_was_string:
+            all_embeddings = all_embeddings[0]
 
         return all_embeddings
 
+
+
+    def start_multi_process_pool(self, target_devices: List[str] = None, encode_batch_size: int = 32):
+        """
+        Starts multi process to process the encode with several, independent  process.
+        This methos is recommend if you want to encode on multiple GPUs. It is advised
+        to start only one process per GPU. This method works together with encode_multi_process
+
+        :param target_devices: PyTorch target devices, e.g. cuda:0, cuda:1... If None, all available CUDA devices will be used
+        :param encode_batch_size: Batch size for each process when calling encode
+        :return: Returns a dict with the target processes, an input queue and and output queue.
+        """
+        if target_devices is None:
+            if torch.cuda.is_available():
+                target_devices = ['cuda:{}'.format(i) for i in range(torch.cuda.device_count())]
+            else:
+                logging.info("CUDA is not available. Start 4 CPU worker")
+                target_devices = ['cpu']*4
+
+        logging.info("Start multi-process pool on devices: {}".format(', '.join(map(str, target_devices))))
+
+        ctx = mp.get_context('spawn')
+        input_queue = ctx.Queue()
+        output_queue = ctx.Queue()
+        processes = []
+
+        for cuda_id in target_devices:
+            p = ctx.Process(target=SentenceTransformer._encode_multi_process_worker, args=(cuda_id, self, input_queue, output_queue, encode_batch_size), daemon=True)
+            p.start()
+            processes.append(p)
+
+        return {'input': input_queue, 'output': output_queue, 'processes': processes}
+
+
+    @staticmethod
+    def stop_multi_process_pool(pool):
+        """
+        Stops all processes started with start_multi_process_pool
+        """
+        for p in pool['processes']:
+            p.terminate()
+
+        for p in pool['processes']:
+            p.join()
+            p.close()
+
+        pool['input'].close()
+        pool['output'].close()
+
+
+    def encode_multi_process(self, sentences: List[str], pool: Dict[str, object], is_pretokenized: bool = False):
+        """
+        This method allows to run encode() on multiple GPUs. The sentences are chunked into smaller packages
+        and sent to individual processes, which encode these on the different GPUs. This method is only suitable
+        for encoding large sets of sentences
+
+        :param sentences: List of sentences
+        :param pool: A pool of workers started with SentenceTransformer.start_multi_process_pool
+        :param is_pretokenized: If true, no tokenization will be applied. It is expected that the input sentences are list of ints.
+        :return: Numpy matrix with all embeddings
+        """
+
+        chunk_size = min(math.ceil(len(sentences) / len(pool["processes"]) / 10), 5000)
+        logging.info("Chunk data into packages of size {}".format(chunk_size))
+
+        if is_pretokenized:
+            sentences_tokenized = sentences
+        else:
+            sentences_tokenized = map(self.tokenize, sentences)
+
+        input_queue = pool['input']
+        num_chunks = 0
+        chunk = []
+
+        for sentence in sentences_tokenized:
+            chunk.append(sentence)
+            if len(chunk) >= chunk_size:
+                input_queue.put([num_chunks, chunk])
+                num_chunks += 1
+                chunk = []
+
+        if len(chunk) > 0:
+            input_queue.put([num_chunks, chunk])
+            num_chunks += 1
+
+        output_queue = pool['output']
+        results_list = sorted([output_queue.get() for _ in range(num_chunks)], key=lambda x: x[0])
+        embeddings = np.concatenate([result[1] for result in results_list])
+        return embeddings
+
+    @staticmethod
+    def _encode_multi_process_worker(target_device: str, model, input_queue, results_queue, encode_batch_size):
+        """
+        Internal working process to encode sentences in multi-process setup
+        """
+        while True:
+            try:
+                id, sentences = input_queue.get()
+                embeddings = model.encode(sentences, device=target_device, is_pretokenized=True, show_progress_bar=False, convert_to_numpy=True, batch_size=encode_batch_size)
+                results_queue.put([id, embeddings])
+            except queue.Empty:
+                break
+
+
     def get_max_seq_length(self):
+        """
+        Returns the maximal sequence length for input the model accepts. Longer inputs will be truncated
+        """
         if hasattr(self._first_module(), 'max_seq_length'):
             return self._first_module().max_seq_length
 
         return None
 
-    def tokenize(self, text):
+    def tokenize(self, text: str):
+        """
+        Tokenizes the text
+        """
         return self._first_module().tokenize(text)
 
     def get_sentence_features(self, *features):
@@ -296,6 +388,33 @@ class SentenceTransformer(nn.Sequential):
         return {'features': features, 'labels': torch.stack(labels)}
 
 
+    def smart_batching_collate_text_only(self, batch):
+        """
+        Transforms a batch from a SmartBatchingDataset to a batch of tensors for the model
+
+        :param batch:
+            a batch from a SmartBatchingDataset
+        :return:
+            a batch of tensors for the model
+        """
+
+        max_seq_len = max([len(text) for text in batch])
+        feature_lists = {}
+
+        for text in batch:
+            sentence_features = self.get_sentence_features(text, max_seq_len)
+            for feature_name in sentence_features:
+                if feature_name not in feature_lists:
+                    feature_lists[feature_name] = []
+
+                feature_lists[feature_name].append(sentence_features[feature_name])
+
+        for feature_name in feature_lists:
+            feature_lists[feature_name] = torch.cat(feature_lists[feature_name])
+
+        return feature_lists
+
+
 
     def fit(self,
             train_objectives: Iterable[Tuple[DataLoader, nn.Module]],
@@ -312,9 +431,7 @@ class SentenceTransformer(nn.Sequential):
             output_path_ignore_not_empty: bool = False,
             save_best_model: bool = True,
             max_grad_norm: float = 1,
-            fp16: bool = False,
-            fp16_opt_level: str = 'O1',
-            local_rank: int = -1
+            use_amp: bool = False
             ):
         """
         Train the model with the given training objective
@@ -323,24 +440,27 @@ class SentenceTransformer(nn.Sequential):
         We sample only as many batches from each objective as there are in the smallest one
         to make sure of equal training with each dataset.
 
-        :param weight_decay:
-        :param scheduler:
-        :param warmup_steps:
-        :param optimizer:
-        :param evaluation_steps:
-        :param output_path:
-        :param output_path_ignore_not_empty: Ignore if the output path contains already files
-        :param save_best_model:
-        :param max_grad_norm:
-        :param fp16:
-        :param fp16_opt_level:
-        :param local_rank:
-        :param train_objectives:
-            Tuples of DataLoader and LossConfig
-        :param evaluator:
-        :param epochs:
-        :param steps_per_epoch: Train for x steps in each epoch. If set to None, the length of the dataset will be used
+        :param train_objectives: Tuples of (DataLoader, LossFunction). Pass more than one for multi-task learning
+        :param evaluator: An evaluator (sentence_transformers.evaluation) evaluates the model performance during training on held-out dev data. It is used to determine the best model that is saved to disc.
+        :param epochs: Number of epochs for training
+        :param steps_per_epoch: Number of training steps per epoch. If set to None (default), one epoch is equal the DataLoader size from train_objectives.
+        :param scheduler: Learning rate scheduler. Available schedulers: constantlr, warmupconstant, warmuplinear, warmupcosine, warmupcosinewithhardrestarts
+        :param warmup_steps: Behavior depends on the scheduler. For WarmupLinear (default), the learning rate is increased from o up to the maximal learning rate. After these many training steps, the learning rate is decreased linearly back to zero.
+        :param optimizer_class: Optimizer
+        :param optimizer_params: Optimizer parameters
+        :param weight_decay: Weight decay for model parameters
+        :param evaluation_steps: If > 0, evaluate the model using evaluator after each number of training steps
+        :param output_path: Storage path for the model and evaluation files
+        :param output_path_ignore_not_empty: By default, training will stop if output_path is not empty. If set to true, this error will be ignored and training proceeds.
+        :param save_best_model: If true, the best model (according to evaluator) is stored at output_path
+        :param max_grad_norm: Used for gradient normalization.
+        :param use_amp: Use Automatic Mixed Precision (AMP). Only for Pytorch >= 1.6.0
         """
+
+        if use_amp:
+            from torch.cuda.amp import autocast
+            scaler = torch.cuda.amp.GradScaler()
+
         self.to(self._target_device)
 
         if output_path is not None:
@@ -379,32 +499,20 @@ class SentenceTransformer(nn.Sequential):
                 {'params': [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)], 'weight_decay': weight_decay},
                 {'params': [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
             ]
-            t_total = num_train_steps
-            if local_rank != -1:
-                t_total = t_total // torch.distributed.get_world_size()
 
             optimizer = optimizer_class(optimizer_grouped_parameters, **optimizer_params)
-            scheduler_obj = self._get_scheduler(optimizer, scheduler=scheduler, warmup_steps=warmup_steps, t_total=t_total)
+            scheduler_obj = self._get_scheduler(optimizer, scheduler=scheduler, warmup_steps=warmup_steps, t_total=num_train_steps)
 
             optimizers.append(optimizer)
             schedulers.append(scheduler_obj)
 
-        if fp16:
-            try:
-                from apex import amp
-            except ImportError:
-                raise ImportError("Please install apex from https://www.github.com/nvidia/apex to use fp16 training.")
-
-            for train_idx in range(len(loss_models)):
-                model, optimizer = amp.initialize(loss_models[train_idx], optimizers[train_idx], opt_level=fp16_opt_level)
-                loss_models[train_idx] = model
-                optimizers[train_idx] = optimizer
 
         global_step = 0
         data_iterators = [iter(dataloader) for dataloader in dataloaders]
 
         num_train_objectives = len(train_objectives)
 
+        skip_scheduler = False
         for epoch in trange(epochs, desc="Epoch"):
             training_steps = 0
 
@@ -428,20 +536,30 @@ class SentenceTransformer(nn.Sequential):
                         data = next(data_iterator)
 
                     features, labels = batch_to_device(data, self._target_device)
-                    loss_value = loss_model(features, labels)
 
-                    if fp16:
-                        with amp.scale_loss(loss_value, optimizer) as scaled_loss:
-                            scaled_loss.backward()
-                        torch.nn.utils.clip_grad_norm_(amp.master_params(optimizer), max_grad_norm)
+                    if use_amp:
+                        with autocast():
+                            loss_value = loss_model(features, labels)
+
+                        scale_before_step = scaler.get_scale()
+                        scaler.scale(loss_value).backward()
+                        scaler.unscale_(optimizer)
+                        torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
+                        scaler.step(optimizer)
+                        scaler.update()
+
+                        skip_scheduler = scaler.get_scale() != scale_before_step
                     else:
+                        loss_value = loss_model(features, labels)
                         loss_value.backward()
                         torch.nn.utils.clip_grad_norm_(loss_model.parameters(), max_grad_norm)
+                        optimizer.step()
 
-                    optimizer.step()
-                    scheduler.step()
                     optimizer.zero_grad()
 
+                if not skip_scheduler:
+                    scheduler.step()
+
                 training_steps += 1
                 global_step += 1
 
@@ -477,7 +595,7 @@ class SentenceTransformer(nn.Sequential):
 
     def _get_scheduler(self, optimizer, scheduler: str, warmup_steps: int, t_total: int):
         """
-        Returns the correct learning rate scheduler
+        Returns the correct learning rate scheduler. Available scheduler: constantlr, warmupconstant, warmuplinear, warmupcosine, warmupcosinewithhardrestarts
         """
         scheduler = scheduler.lower()
         if scheduler == 'constantlr':
@@ -523,4 +641,18 @@ class SentenceTransformer(nn.Sequential):
         """
         Property to set the tokenizer that is should used by this model
         """
-        self._first_module().tokenizer = value
\ No newline at end of file
+        self._first_module().tokenizer = value
+
+    @property
+    def max_seq_length(self):
+        """
+        Property to get the maximal input sequence length for the model. Longer inputs will be truncated.
+        """
+        return self._first_module().max_seq_length
+
+    @max_seq_length.setter
+    def max_seq_length(self, value):
+        """
+        Property to set the maximal input sequence length for the model. Longer inputs will be truncated.
+        """
+        self._first_module().max_seq_length = value
\ No newline at end of file
